﻿#region GPL License

//Mediaportal (http://www.team-mediaportal.com/) Plugin - WatchedSynchronizer
//Copyright (C) 2015 Max Wimmelbacher
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Collections.ObjectModel;
using MediaPortal.GUI.Library;
using MediaPortal.Dialogs;
using MediaPortal.Player;
using MediaPortal.Video.Database;
using MediaPortal.Configuration;
using MediaPortal.Profile;
using WindowPlugins.GUITVSeries;
using MediaPortal.Plugins.MovingPictures.MainUI;
using MediaPortal.Plugins.MovingPictures.Database;
using WatchedSynchronizer.Configuration;
using System.Collections.Concurrent;


namespace WatchedSynchronizer
{
  [PluginIcons("WatchedSynchronizer.Resources.WatchedSynchronizerEnabled.png", "WatchedSynchronizer.Resources.WatchedSynchronizerDisabled.png")]

  public class WatchedSynchronizer : ISetupForm, IPlugin
  {

    #region Enumerations

    /// <summary>
    /// Enumeration of media types which are used to determine the correct databases to be updated.
    /// </summary>

    public enum MediaType
    {
      MPTVSeries,
      MPVideo,
      MovingPictures,
      Unknown
    };

    public enum EventType
    {
        StopOrChanged,
        Ended,
        Toggle
    };

    #endregion

    #region Declaration

    /// <summary>
    /// Instantiate attributes on object construction.
    /// mCurrentMediaType is set to "Unknown" because on plugin startup no playback should be running.
    /// </summary>

    private WatchedSynchronizerConfiguration mConfiguration;
    private Collection<AdditionalVideoDatabase> mVideoDatabases = new Collection<AdditionalVideoDatabase>();
    private Collection<AdditionalTVSeriesDatabase> mTVSeriesDatabases = new Collection<AdditionalTVSeriesDatabase>();
    private Collection<AdditionalMovingPicturesDatabase> mMovingPicturesDatabases = new Collection<AdditionalMovingPicturesDatabase>();
    private static int mWatchedPercentageVideoDatabase;
    private static bool mMarkWatchedFilesVideoDatabase;
    private static int mWatchedPercentageTVSeriesDatabase;
    private WatchedSynchronizer.MediaType mCurrentMediaType = MediaType.Unknown;
    static ConcurrentQueue<WatchedStateEvent> WatchStateEventQueue = new ConcurrentQueue<WatchedStateEvent>();
    // Create the thread object. This does not start the thread.
    WorkerThread workerObject = new WorkerThread();
    Thread workerThread = null;

    public struct gPlayer
    {
        public int SetResumeBDTitleState;
        public double CurrentPosition;
        public string CurrentFile;
        public double Duration;
        public bool IsDVDMenu;
        public byte[] bteResumeData;
    };

    public struct WatchedStateEvent
    {
        public EventType EventType;
        public DateTime dtNextCheck;
        public object objDB;
        public MediaType MediaType;
        public string strFileName;
        public int intStopTime;
        public bool bolWatched;
        public bool bolUse_gPlayer;
        public gPlayer gPlayer;
    };

    #endregion

    #region ISetupForm Members

    /// <summary>
    /// This region is providing the necessary attributes/methods required by Mediaportal for the plugin configuration.
    /// </summary>
    
    public string Author()
    {
      return "Max Wimmelbacher";
    }

    public bool CanEnable()
    {
      return true;
    }

    public string Description()
    {
      return "Process plugin to update the watched information for media across several databases.";
    }

    public bool DefaultEnabled()
    {
      return false;
    }

    public bool GetHome(out string strButtonText, out string strButtonImage, out string strButtonImageFocus, out string strPictureImage)
    {
      strButtonText = string.Empty;
      strButtonImage = string.Empty;
      strButtonImageFocus = string.Empty;
      strPictureImage = string.Empty;
      return false;
    }

    public int GetWindowId()
    {
      return -1;
    }

    public bool HasSetup()
    {
      return true;
    }

    public string PluginName()
    {
      return "WatchedSynchronizer";
    }

    public void ShowPlugin()
    {
      var Config = new WatchedSynchronizerConfig();
      Config.Show();
    }

    #endregion
   
    #region IPlugin Members

    /// <summary>
    /// This region is providing the necessary methods required by Mediaportal for interaction with the plugin.
    /// Configuration Loading and hooking on the events necessary to trigger plugin actions is done here.
    /// </summary>
    
    public void Start()
    {
      Log.Info("WatchedSynchronizer: Starting");
      LoadConfiguration();
      GUIWindowManager.Receivers += new SendMessageHandler(OnGUIMessageReceived);
      TVSeriesPlugin.ToggleWatched += new TVSeriesPlugin.ToggleWatchedEventDelegate(OnTVSeriesToggledWatched);
      MovingPicturesGUI.ToggleWatched += new MovingPicturesGUI.ToggleWatchedEventDelegate(OnMovingPicturesToggledWatched);
      g_Player.PlayBackEnded += OnPlayBackEnded;
      g_Player.PlayBackChanged += OnPlayBackChanged;
      g_Player.PlayBackStopped += OnPlayBackStopped;


      //Create the workerthread object
      workerThread = new Thread(workerObject.DoWork);

      // Start the worker thread.
      workerThread.Start();
    }

    public void Stop()
    {
      Log.Info("WatchedSynchronizer: Stopping");
      GUIWindowManager.Receivers -= new SendMessageHandler(OnGUIMessageReceived);
      TVSeriesPlugin.ToggleWatched -= new TVSeriesPlugin.ToggleWatchedEventDelegate(OnTVSeriesToggledWatched);
      MovingPicturesGUI.ToggleWatched -= new MovingPicturesGUI.ToggleWatchedEventDelegate(OnMovingPicturesToggledWatched);
      g_Player.PlayBackEnded -= OnPlayBackEnded;
      g_Player.PlayBackChanged -= OnPlayBackChanged;
      g_Player.PlayBackStopped -= OnPlayBackStopped;

      // Request that the worker thread stop itself:
      workerObject.RequestStop();

      // Use the Join method to block the current thread  
      // until the object's thread terminates.
      workerThread.Join();
      Log.Info("WatchedSynchronizer: Stopped!");
    }

    #endregion

    #region Mediaportal.GUI.Library.GUIWindowManager events

    /// <summary>
    /// The method OnGUIMessageReceived is listening for messages indicating that a media playback started and by whom.
    /// This information is used to determine which database needs to be updated after the media playback stopped.
    /// </summary>

    private void OnGUIMessageReceived(GUIMessage objMessage)
    {
      if (objMessage.Message == GUIMessage.MessageType.GUI_MSG_PLAYBACK_STARTED)
      {
        Log.Debug("WatchedSynchronizer: Encountered message 'GUI_MSG_PLAYBACK_STARTED' for WindowId '" + GUIWindowManager.GetPreviousActiveWindow() + "'.");
        switch (GUIWindowManager.GetPreviousActiveWindow())
        {
          case 25:
          case 2003:
            mCurrentMediaType = MediaType.MPVideo;
            break;

          case 2005:
            mCurrentMediaType = MediaType.MovingPictures;
            break;

          case 9811:
            mCurrentMediaType = MediaType.MPTVSeries;
            break;
          default:
            {
              mCurrentMediaType = MediaType.Unknown;
              Log.Warn("WatchedSynchronizer: Encountered WindowId '" + GUIWindowManager.GetPreviousActiveWindow() + "' which has no mediatype assigned.");
              break;
            }
        }
      }
    }

    #endregion

    #region MediaPortal.Player.g_player events

    /// <summary>
    /// This region defindes the methods to be run on triggering of hooked events.
    /// Based on the object which started the media playback it is decided which update method needs to be called.
    /// </summary>

    private void OnPlayBackEnded(g_Player.MediaType enuType, string strFileName)
    {
      WatchedStateEvent curEvent = new WatchedStateEvent();

      if (enuType == g_Player.MediaType.Video)
      {
        Log.Debug("WatchedSynchronizer: 'g_Player.PlayBackEnded' event with media type '" + mCurrentMediaType + "' occured for file '" + strFileName + "'.");

        //Create Watched Event
        curEvent.EventType = EventType.Ended;
        curEvent.MediaType = mCurrentMediaType;
        curEvent.strFileName = strFileName;
        curEvent.intStopTime = 0;
        curEvent.bolWatched = false;
        curEvent.bolUse_gPlayer = true;

        //Enqueue Event
        FireWatchedEvent(curEvent);
      }
    }

    private void OnPlayBackChanged(g_Player.MediaType enuType, int intStopTime, string strFileName)
    {
        WatchedStateEvent curEvent = new WatchedStateEvent();

        if (enuType == g_Player.MediaType.Video)
        {
            Log.Debug("WatchedSynchronizer: 'g_Player.PlayBackChanged' event with media type '" + mCurrentMediaType + "' occured for file '" + strFileName + "'.");

            //Create Watched Event
            curEvent.EventType = EventType.StopOrChanged;
            curEvent.MediaType = mCurrentMediaType;
            curEvent.strFileName = strFileName;
            curEvent.intStopTime = intStopTime;
            curEvent.bolWatched = false;
            curEvent.bolUse_gPlayer = true;

            //Enqueue Event
            FireWatchedEvent(curEvent);
        }
    }

    private void OnPlayBackStopped(g_Player.MediaType enuType, int intStopTime, string strFileName)
    {
        WatchedStateEvent curEvent = new WatchedStateEvent();

      if (enuType == g_Player.MediaType.Video)
      {
        Log.Debug("WatchedSynchronizer: 'g_Player.PlayBackStopped' event with media type '" + mCurrentMediaType + "' occured for file '" + strFileName + "'.");

        //Create Watched Event
        curEvent.EventType = EventType.StopOrChanged;
        curEvent.MediaType = mCurrentMediaType;
        curEvent.strFileName = strFileName;
        curEvent.intStopTime = intStopTime;
        curEvent.bolWatched = false;
        curEvent.bolUse_gPlayer = true;

        //Enqueue Event
        FireWatchedEvent(curEvent);
      }
    }

    #endregion

    #region WindowPlugins.GUITVSeries.TVSeriesPlugin events

    /// <summary>
    /// The method OnTVSeriesToggledWatched is listening for messages indicating that a episod was toggled watched using the menu in MediaPortal.
    /// This information is used to set the information related to Watched
    /// </summary>
    /// 

    private void OnTVSeriesToggledWatched(DBSeries objSeries, List<DBEpisode> lstDBEpisodes, bool bolWatched)
    {
      WatchedStateEvent curEvent = new WatchedStateEvent();

      foreach (DBEpisode objLoop in lstDBEpisodes)
      {
          Log.Debug("WatchedSynchronizer: OnTVSeriesToggledWatched: The Watched Status for file '" + objLoop[DBEpisode.cFilename] + "' was toggled to '" + bolWatched + "',");

        //Create Watched Event
        curEvent.EventType = EventType.Toggle;
        curEvent.MediaType = MediaType.MPTVSeries;
        curEvent.strFileName = objLoop[DBEpisode.cFilename];
        curEvent.intStopTime = 0;
        curEvent.bolWatched = bolWatched;
        curEvent.bolUse_gPlayer = false;

        //Enqueue Event
        FireWatchedEvent(curEvent);
      }
    }

    #endregion

    #region MediaPortal.Plugins.MovingPictures.MainUI.MovingPicturesGUI events

    /// <summary>
    /// The method OnMovingPIcturesToggledWatched is listening for messages indicating that a movie was toggled watched using the menu in MediaPortal.
    /// This information is used to set the information related to Watched
    /// </summary>
    /// 

    private void OnMovingPicturesToggledWatched(DBMovieInfo movie, bool bolWatched)
    {
        WatchedStateEvent curEvent = new WatchedStateEvent();
        DBLocalMedia localMedia = movie.LocalMedia[0];

        Log.Debug("WatchedSynchronizer: OnMovingPicturesToggledWatched: The Watched Status for file '" + localMedia.FullPath + "' was toggled to '" + bolWatched + "',");

        //Create Watched Event
        curEvent.EventType = EventType.Toggle;
        curEvent.MediaType = MediaType.MovingPictures;
        curEvent.strFileName = localMedia.FullPath;
        curEvent.intStopTime = 0;
        curEvent.bolWatched = bolWatched;
        curEvent.bolUse_gPlayer = false;

        //Enqueue Event
        FireWatchedEvent(curEvent);
    }

    #endregion

    #region WorkerThread

    public class WorkerThread
    {
        // This method will be called when the thread is started. 
        public void DoWork()
        {
            WatchedStateEvent evt;
            int iSleepMs = 10, iEventCount = 0, res;
            TimeSpan span;

            Log.Debug("WatchedSynchronizer: WorkerThread started!");
            while (!_shouldStop)
            {
                //SleepTime when there are no events

                while (WatchStateEventQueue.TryDequeue(out evt))
                {
                    iEventCount++;
                    if (evt.dtNextCheck <= DateTime.Now)
                    {
                        //we should process this item now
                        Log.Debug("WatchedSynchronizer: WorkerThread: New event to process: " +
                                    "EventType: '" + evt.EventType.ToString() + "' - " +
                                    "dtNextCheck: '" + evt.dtNextCheck.ToString("HH:mm:ss.fff tt") + "' - " +
                                    "MediaType: '" + evt.MediaType.ToString() + "' - " +
                                    "strFileName: '" + evt.strFileName + "' - " +
                                    "intStopTime: '" + evt.intStopTime.ToString() + "' - " +
                                    "bolWatched: '" + evt.bolWatched.ToString() + "'"
                                    );
                        if (evt.bolUse_gPlayer)
                        {
                            Log.Debug("WatchedSynchronizer: WorkerThread: gPlayer data:" +

                                        "gPlayer.SetResumeBDTitleState: '" + evt.gPlayer.SetResumeBDTitleState.ToString() + "' - " +
                                        "gPlayer.CurrentPosition: '" + evt.gPlayer.CurrentPosition.ToString() + "' - " +
                                        "gPlayer.CurrentFile: '" + evt.gPlayer.CurrentFile.ToString() + "' - " +
                                        "gPlayer.Duration: '" + evt.gPlayer.Duration.ToString() + "' - " +
                                        "gPlayer.IsDVDMenu: '" + evt.gPlayer.IsDVDMenu.ToString() + "'" /*+
                                        "gPlayer.bteResumeData: '" + evt.gPlayer.bteResumeData.ToString()*/
                                        );
                        }


                        //Process the watchedevent
                        res = 0;
                        switch (evt.MediaType)
                        {
                            case MediaType.MPTVSeries:
                                res = ProcessMPTVSeriesWatchedStateEvent(evt);
                                break;

                            case MediaType.MovingPictures:
                                Log.Debug("WatchedSynchronizer: WorkerThread: TODO: implement ProcessMovingPicturesWatchedStateEvent!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                                //res = ProcessMovingPicturesWatchedStateEvent(evt);
                                break;

                            case MediaType.MPVideo:
                                res = ProcessMPVideoWatchedStateEvent(evt);
                                break;

                            default:
                                Log.Debug("WatchedSynchronizer: WorkerThread: dropped event with unknown MediaType!");
                                break;


                        }

                        //check result and reschedule/retry if needed
                        if (res < 0)
                        {
                            //We should reschedule/retray the event
                            evt.dtNextCheck = DateTime.Now.AddSeconds(5);   //Retry in 5 Seconds
                            WatchStateEventQueue.Enqueue(evt);  //Event should be resheduled
                        }

                    }
                    else
                    {
                        //Event should be resheduled, readd it to queue
                        WatchStateEventQueue.Enqueue(evt);
                    }

                    //calculate complete millisconds to sleep
                    if(evt.dtNextCheck > DateTime.Now)
                    {
                        span = evt.dtNextCheck - DateTime.Now;
                        iSleepMs = (int)span.TotalMilliseconds;
                    }
                }

                //Minimal sleep is 10ms to give other proceses cpu time
                if (iEventCount == 0) iSleepMs = 100;       //SleepTime when there are no events
                if (iSleepMs < 10) iSleepMs = 10;
                Thread.Sleep(iSleepMs);
            }
            Log.Debug("WatchedSynchronizer: WorkerThread: terminating gracefully.");
        }
        public void RequestStop()
        {
            _shouldStop = true;
        }
        // Volatile is used as hint to the compiler that this data 
        // member will be accessed by multiple threads. 
        private volatile bool _shouldStop;
    }

    #endregion

    #region Private methods

    /// <summary>
    /// The method LoadConfiguration is called on plugin start and is deserializing the saved configuration.
    /// Database objects are created based on the configuration and stored in collections for the process lifetime.
    /// </summary>

    private void LoadConfiguration()
    {

      //Loading the plugins configuration.

      MPSettings objSettings = MPSettings.Instance;
      string strConfiguration = objSettings.GetValueAsString("watchedsynchronizer", "configuration", string.Empty);
      if (string.IsNullOrEmpty(strConfiguration))
      {
        mConfiguration.mConfigurationEntries = new List<WatchedSynchronizerConfigurationEntry>();
      }
      else
      {
        mConfiguration = Utils.Deserialize(strConfiguration);
      }

      //Loading the configured values when a media item is considered watched.

      mWatchedPercentageVideoDatabase = objSettings.GetValueAsInt("movies", "playedpercentagewatched", 95);
      mMarkWatchedFilesVideoDatabase = objSettings.GetValueAsBool("movies", "markwatched", true);
      mWatchedPercentageTVSeriesDatabase = (int)DBOption.GetOptions(DBOption.cWatchedAfter);

      //Create object collections containing all configured databases which needs to be updated.

      foreach (WatchedSynchronizerConfigurationEntry objLoop in mConfiguration.mConfigurationEntries)
      {
        if (objLoop.mEnabled == true)
        {
          switch (objLoop.mDatabaseType)
          {
            case "Video database":
              AdditionalVideoDatabase objVideoDatabase = new AdditionalVideoDatabase(objLoop.mDatabasePath);
              objVideoDatabase.Dispose();
              mVideoDatabases.Add(objVideoDatabase);
              break;
            case "TVSeries database":
              AdditionalTVSeriesDatabase objTVSeriesDatabase = new AdditionalTVSeriesDatabase(objLoop.mDatabasePath);
              objTVSeriesDatabase.Dispose();
              mTVSeriesDatabases.Add(objTVSeriesDatabase);
              break;
            case "MovingPictures database":
              AdditionalMovingPicturesDatabase objMovingPicturesDatabase = new AdditionalMovingPicturesDatabase(objLoop.mDatabasePath);
              objMovingPicturesDatabase.Dispose();
              mMovingPicturesDatabases.Add(objMovingPicturesDatabase);
              break;
          }
        }
      }
    }


    //Method FireWatchedEvent will be triggerd on any Mediaportal Stop, change, toggle event and adds the needed data to the WatchStateEventQueue
    private void FireWatchedEvent(WatchedStateEvent evt)
    {
        //Add needed g_Player data if evt.bolUse_gPlayer is set
        if (evt.bolUse_gPlayer)
        {
            try
            {
                evt.gPlayer.SetResumeBDTitleState = g_Player.SetResumeBDTitleState;
                evt.gPlayer.CurrentPosition = g_Player.Player.CurrentPosition;
                evt.gPlayer.CurrentFile = g_Player.CurrentFile;
                evt.gPlayer.Duration = g_Player.Player.Duration;
                evt.gPlayer.IsDVDMenu = g_Player.IsDVDMenu;
                g_Player.Player.GetResumeState(out evt.gPlayer.bteResumeData);
            }
            catch (Exception ex)
            {
                Log.Error("WatchedSynchronizer: FireWatchedEvent:exception err:{0} stack:{1}", ex.Message, ex.StackTrace);
            }
        }
        else
        {
            evt.gPlayer.SetResumeBDTitleState = 0;
            evt.gPlayer.CurrentPosition = 0;
            evt.gPlayer.CurrentFile = "";
            evt.gPlayer.Duration = 0;
            evt.gPlayer.IsDVDMenu = false;
            evt.gPlayer.bteResumeData = new byte[] {0x00 };
        }

        //DateTime when the item should be processed
        evt.dtNextCheck = DateTime.Now;

        //Enqueue Event for every configured database
        switch (evt.MediaType)
        {
            case MediaType.MPTVSeries:
                foreach (AdditionalTVSeriesDatabase objLoop in mTVSeriesDatabases)
                {
                    evt.objDB = objLoop;
                    WatchStateEventQueue.Enqueue(evt);
                }
                break;

            case MediaType.MovingPictures:
                foreach (AdditionalMovingPicturesDatabase objLoop in mMovingPicturesDatabases)
                {
                    evt.objDB = objLoop;
                    WatchStateEventQueue.Enqueue(evt);
                }
                break;

            case MediaType.MPVideo:
                foreach (AdditionalVideoDatabase objLoop in mVideoDatabases)
                {
                    evt.objDB = objLoop;
                    WatchStateEventQueue.Enqueue(evt);
                }
                break;

            default:
                Log.Debug("WatchedSynchronizer: FireWatchedEvent: unknown MediaType");
                break;


        }

    }

    #region Video databases

    /// <summary>
    /// Methods in this region are run if the media type is "MPVideo" and a gplayer event was triggered.
    /// The defined methods run the necessary commands to update all configured video databases.
    /// </summary>

    //Method ProcessMPVideoWatchedStateEvent is run when an event for videos databases should be processed
    //returns:
    //  1 = Infos written to given DB -> event could be cleared from queue
    //  0 = other error (ex. file not found in db, file does not exists...) -> event could be cleared from queue
    //  <0 = DB is locked by other user -> event should NOT be cleared from queue -> retry


    private static int ProcessMPVideoWatchedStateEvent(WatchedStateEvent evt)
    {
        bool bolStacked = false;
        AdditionalVideoDatabase MPVideoDB = (AdditionalVideoDatabase)evt.objDB;

        MPVideoDB.ReOpen();
        bool bolUpdated = false;
        ArrayList alsFiles = new ArrayList();
        int intDuration = 0;
        int intTotalDuration = 0;
        int intPlayTimePercentage = 0;

        //Get MovieID and all its correlated media file(s).

        int intMovieId = MPVideoDB.GetMovieId(evt.strFileName);
        if (intMovieId == -1)
        {
            Log.Warn("WatchedSynchronizer: ProcessMPVideoWatchedStateEvent: File '" + evt.strFileName + "' is not available in database '" + MPVideoDB.DatabaseName + "'.");
            MPVideoDB.Dispose();
            return 0;
        }
        else
        {
            MPVideoDB.GetFilesForMovie(intMovieId, ref alsFiles);
        }

        //Check if MovieID consists of several media files.

        if (alsFiles.Count <= 0)
        {
            Log.Warn("WatchedSynchronizer: ProcessMPVideoWatchedStateEvent: File '" + evt.strFileName + "' is not available in database '" + MPVideoDB.DatabaseName + "'.");
            MPVideoDB.Dispose();
            return 0;
        }
        else if (alsFiles.Count > 1)
        {
            bolStacked = true;
        }

        //Calculates the total duration of all media files of the MovieID.

        foreach (string strLoop in alsFiles)
        {
            intTotalDuration += MPVideoDB.GetVideoDuration(MPVideoDB.GetFileId(strLoop));
        }

        //do calculations depending on the EventType
        switch(evt.EventType)
        {
        
            case EventType.Ended:

                //Resets the MovieStopTime and the ResumeData for the media file which was last played.

                for (int intLoop = 0; intLoop < alsFiles.Count; intLoop++)
                {
                    string strFilePath = (string)alsFiles[intLoop];
                    byte[] bteResumeData = null;
                    int intFileId = MPVideoDB.GetFileId(strFilePath);
                    if (intFileId < 0)
                    {
                        break;
                    }
                    MPVideoDB.GetMovieStopTimeAndResumeData(intFileId, out bteResumeData, evt.gPlayer.SetResumeBDTitleState);
                    MPVideoDB.SetMovieStopTimeAndResumeData(intFileId, 0, bteResumeData, evt.gPlayer.SetResumeBDTitleState);
                    bolUpdated = true;
                }

                //Calculates the Watched percentage of the movie based on all media items correlated.

                if (bolStacked && intTotalDuration != 0)
                {
                    for (int i = 0; i < alsFiles.Count; i++)
                    {
                        int intFileId = MPVideoDB.GetFileId((string)alsFiles[i]);
                        if (evt.strFileName != (string)alsFiles[i])
                        {
                            intDuration += MPVideoDB.GetVideoDuration(intFileId);
                            continue;
                        }
                        intPlayTimePercentage = (int)(100 * (intDuration + evt.gPlayer.CurrentPosition) / intTotalDuration);
                        break;
                    }
                }
                else
                {
                    intPlayTimePercentage = 100;
                }

                break;

           case EventType.StopOrChanged:

                //Calculates the Watched percentage of the movie based on all media items correlated.
                if (bolStacked && intTotalDuration != 0)
                {
                    for (int intLoop = 0; intLoop < alsFiles.Count; intLoop++)
                    {
                        if (evt.gPlayer.CurrentFile != (string)alsFiles[intLoop])
                        {
                            intDuration += MPVideoDB.GetVideoDuration(MPVideoDB.GetFileId((string)alsFiles[intLoop]));
                            continue;
                        }
                        intPlayTimePercentage = (100 * (intDuration + evt.intStopTime) / intTotalDuration);
                        break;
                    }
                }
                else
                {
                    if (evt.gPlayer.Duration >= 1)
                    {
                        intPlayTimePercentage = (int)Math.Ceiling((evt.intStopTime / evt.gPlayer.Duration) * 100);
                    }
                }

                //Sets the Watched status and correlated for the media file which was last played.

                for (int intLoop = 0; intLoop < alsFiles.Count; intLoop++)
                {
                    string strFilePath = (string)alsFiles[intLoop];
                    int intFileId = MPVideoDB.GetFileId(strFilePath);
                    intMovieId = MPVideoDB.GetMovieId(strFilePath);

                    //Check if a DVD/Blueray was played and set the Watched status and correlated according to that.

                    if (evt.gPlayer.IsDVDMenu)
                    {
                        MPVideoDB.SetMovieStopTimeAndResumeData(intFileId, 0, null, evt.gPlayer.SetResumeBDTitleState);
                        MPVideoDB.SetMovieWatchedStatus(intMovieId, true, 100);
                        MPVideoDB.MovieWatchedCountIncrease(intMovieId);
                    }

                    //Check if the media file currently played is the one in the loop and set the Watched status and correlated according to that.

                    else if ((evt.strFileName.Trim().ToLowerInvariant().Equals(strFilePath.Trim().ToLowerInvariant())) && (evt.intStopTime > 0))
                    {
                        MPVideoDB.GetMovieStopTimeAndResumeData(intFileId, out evt.gPlayer.bteResumeData, evt.gPlayer.SetResumeBDTitleState);
                        MPVideoDB.SetMovieStopTimeAndResumeData(intFileId, evt.intStopTime, evt.gPlayer.bteResumeData, evt.gPlayer.SetResumeBDTitleState);

                        //Updates the Watched status for all media files based on the configuration done for the video database.

                        if (intPlayTimePercentage >= mWatchedPercentageVideoDatabase)
                        {
                            MPVideoDB.SetMovieWatchedStatus(intMovieId, true, intPlayTimePercentage);
                            MPVideoDB.MovieWatchedCountIncrease(intMovieId);
                            bolUpdated = true;
                        }
                        else
                        {
                            int intPercent = 0;
                            int intTimesWatched = 0;
                            bool bolWatched = MPVideoDB.GetMovieWatchedStatus(intMovieId, out intPercent, out intTimesWatched);
                            MPVideoDB.SetMovieWatchedStatus(intMovieId, bolWatched, intPlayTimePercentage);
                            bolUpdated = true;
                        }
                    }
                    else
                    {
                        MPVideoDB.DeleteMovieStopTime(intFileId);
                    }
                }

                break;

           case EventType.Toggle:
           default:
                Log.Debug("WatchedSynchronizer: ProcessMPVideoWatchedStateEvent: unkown EventType: " + evt.EventType.ToString());
                break;

        }



        //Updates the Watched status based on the configuration done for the video database.
        if (mMarkWatchedFilesVideoDatabase)
        {
            IMDBMovie objIMDBMovie = new IMDBMovie();
            MPVideoDB.GetMovieInfoById(intMovieId, ref objIMDBMovie);
            if (!objIMDBMovie.IsEmpty)
            {

                //Updates the Watched status for all media files based on the configuration done for the video database. 
                if (intPlayTimePercentage >= mWatchedPercentageVideoDatabase)
                {
                    objIMDBMovie.Watched = 1;

                    if(evt.EventType == EventType.Ended)
                    {
                        MPVideoDB.SetWatched(objIMDBMovie);
                        MPVideoDB.SetMovieWatchedStatus(intMovieId, true, intPlayTimePercentage);
                        MPVideoDB.MovieWatchedCountIncrease(intMovieId);
                        MPVideoDB.SetDateWatched(objIMDBMovie);
                    }else if(evt.EventType == EventType.Ended)
                    {
                        objIMDBMovie.DateWatched = DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss");
                        MPVideoDB.SetMovieInfoById(intMovieId, ref objIMDBMovie);
                    }
                    bolUpdated = true;
                }
                else if(evt.EventType == EventType.Ended)
                {
                    int intPercent = 0;
                    int intTimesWatched = 0;
                    bool bolWatched = MPVideoDB.GetMovieWatchedStatus(intMovieId, out intPercent, out intTimesWatched);
                    MPVideoDB.SetMovieWatchedStatus(intMovieId, bolWatched, intPlayTimePercentage);
                    bolUpdated = true;
                }



            }
        }

        //Check if update was successfull and write to media portal log.

        if (bolUpdated == true)
        {
            Log.Info("WatchedSynchronizer: ProcessMPVideoWatchedStateEvent: Information for file '" + evt.strFileName + "' was updated in database '" + MPVideoDB.DatabaseName + "'.");
        }
        MPVideoDB.Dispose();

        //OK, clear event from queue
        return 1;
    }



    #endregion


    #region TVSeries databases

    /// <summary>
    /// Methods in this region are run if the media type is "MPTVSeries" or the trigger was an event related to "MPTVSeries".
    /// The defined methods run the necessary commands to update all configured video databases.
    /// </summary>


    //Method ProcessMPVideoWatchedStateEvent is run when an event for videos databases should be processed
    //returns:
    //  1 = Infos written to given DB -> event could be cleared from queue
    //  0 = other error (ex. file not found in db, file does not exists...) -> event could be cleared from queue
    //  <0 = DB is locked by other user -> event should NOT be cleared from queue -> retry

    private static int ProcessMPTVSeriesWatchedStateEvent(WatchedStateEvent evt)
    {
        int intPlayTimePercentage = 0;
        AdditionalTVSeriesDatabase TVSeriesDB = (AdditionalTVSeriesDatabase)evt.objDB;

        TVSeriesDB.ReOpen();
        bool bolUpdated = false;
        ArrayList alsCompositeIds = new ArrayList();

        //Get CompositeId(s).

        string strCompositeId = TVSeriesDB.GetCompositeId(evt.strFileName);
        if (strCompositeId == string.Empty)
        {
            Log.Warn("WatchedSynchronizer: File '" + evt.strFileName + "' is not available in database '" + TVSeriesDB.DatabaseName + "'.");
            TVSeriesDB.Dispose();
            return 0;
        }
        else
        {
            TVSeriesDB.GetCompositeIdsForEpisode(strCompositeId, ref alsCompositeIds);
        }
        if (alsCompositeIds.Count <= 0)
        {
            Log.Warn("WatchedSynchronizer: File '" + evt.strFileName + "' is not available in database '" + TVSeriesDB.DatabaseName + "'.");
            TVSeriesDB.Dispose();
            return 0;
        }

        //Calculates the Watched percentage of the episode.

        if ((evt.gPlayer.Duration >= 1) && (evt.EventType == EventType.StopOrChanged))
        {
            intPlayTimePercentage = (int)Math.Ceiling((evt.intStopTime / evt.gPlayer.Duration) * 100);
        }
        else if (evt.EventType == EventType.Ended)
        {
            intPlayTimePercentage = 100;
        }

        //Updates the Watched status for all CompositeIds.

        for (int intLoop = 0; intLoop < alsCompositeIds.Count; intLoop++)
        {
            strCompositeId = (string)alsCompositeIds[intLoop];
            bool bolWatched = false;
            string strWatchedDate;
            int intStopTime;

            //Check if episode was already watched.

            TVSeriesDB.GetEpisodeWatchedStatus(strCompositeId, out bolWatched, out strWatchedDate, out intStopTime);

            //Updates the Watched status based on the configuration done for the tvseries database

            if (((evt.EventType != EventType.Toggle) && (intPlayTimePercentage >= mWatchedPercentageTVSeriesDatabase)) || ((evt.EventType == EventType.Toggle) && (bolWatched == false && evt.bolWatched == true)))
            {
                if (((evt.EventType != EventType.Toggle) && (bolWatched == false)) || ((evt.EventType == EventType.Toggle) && (bolWatched == false && evt.bolWatched == true)))
                {

                    //Update the Watched status for the episodes and correlated it is unwatched.

                    TVSeriesDB.SetEpisodeWatchedStatus(strCompositeId, true, DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss"), 0, true);
                    if (evt.EventType == EventType.Toggle) TVSeriesDB.EpisodePlayCountIncrease(strCompositeId, true);
                    string strSeasonId;
                    bool bolSeasonWatched = false;
                    TVSeriesDB.GetSeasonWatchedStatus(strCompositeId, out strSeasonId, out bolSeasonWatched);
                    if (bolSeasonWatched == false)
                    {

                        //Update the Unwatched episodes count for the season if it is unwatched.

                        TVSeriesDB.SeasonUnWatchedCountDecrease(strSeasonId);
                        string strSeriesId;
                        bool bolSeriesWatched = false;
                        TVSeriesDB.GetSeriesWatchedStatus(strCompositeId, out strSeriesId, out bolSeriesWatched);
                        if (bolSeriesWatched == false)
                        {

                            //Update the Unwatched episodes count for the series if it is unwatched.

                            TVSeriesDB.SeriesUnWatchedCountDecrease(strSeriesId);
                        }
                    }
                }
                else
                {
                    if (evt.EventType == EventType.StopOrChanged) strWatchedDate = DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss");
                    TVSeriesDB.SetEpisodeWatchedStatus(strCompositeId, bolWatched, strWatchedDate, 0);
                }
                TVSeriesDB.EpisodePlayCountIncrease(strCompositeId);
                bolUpdated = true;
            }
            else
            {
                if (evt.EventType == EventType.Ended) intStopTime = 0;
                if (evt.EventType == EventType.StopOrChanged) intStopTime = evt.intStopTime;
                if (evt.EventType == EventType.Toggle) bolWatched = evt.bolWatched;

                //Update Watched State
                TVSeriesDB.SetEpisodeWatchedStatus(strCompositeId, bolWatched, strWatchedDate, intStopTime);

                if (evt.EventType == EventType.Ended) TVSeriesDB.EpisodePlayCountIncrease(strCompositeId);
                bolUpdated = true;
            }
        }

        //Check if update was successfull and write to media portal log.

        if (bolUpdated == true)
        {
            Log.Info("WatchedSynchronizer: Information for file '" + evt.strFileName + "' was updated in database '" + TVSeriesDB.DatabaseName + "'.");
        }
        TVSeriesDB.Dispose();

        //OK, clear event from queue
        return 1;
    }

    #endregion

    #endregion


  }
}